@@ -39,13 +39,13 @@ import org.jruby.RubyBoolean;
 import org.jruby.RubyClass;
 import org.jruby.RubyFixnum;
 import org.jruby.RubyHash;
+import org.jruby.RubyInteger;
 import org.jruby.RubyMatchData;
 import org.jruby.RubyNumeric;
 import org.jruby.RubyObject;
 import org.jruby.RubyRegexp;
 import org.jruby.RubyString;
 import org.jruby.RubySymbol;
-import org.jruby.RubyThread;
 import org.jruby.anno.JRubyClass;
 import org.jruby.anno.JRubyMethod;
 import org.jruby.ast.util.ArgsUtil;
@@ -55,6 +55,7 @@ import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.util.ByteList;
 import org.jruby.util.StringSupport;
+import org.jruby.util.ConvertBytes;
 
 import java.util.Iterator;
 
@@ -69,6 +70,8 @@ import static org.jruby.runtime.Visibility.PRIVATE;
 public class RubyStringScanner extends RubyObject {
     private static final long serialVersionUID = -3722138049229128675L;
 
+    private static final String STRSCAN_VERSION = "3.1.2";
+
     private RubyString str;
     private int curr = 0;
     private int prev = -1;
@@ -78,8 +81,6 @@ public class RubyStringScanner extends RubyObject {
     private boolean matched;
     private boolean fixedAnchor;
 
-    private static final int MATCHED_STR_SCN_F = 1 << 11;
-
     public static RubyClass createScannerClass(final Ruby runtime) {
         RubyClass Object = runtime.getObject();
 
@@ -91,7 +92,7 @@ public class RubyStringScanner extends RubyObject {
             Object.defineConstant("ScanError", error);
         }
 
-        RubyString version = runtime.newString("3.0.2");
+        RubyString version = runtime.newString(STRSCAN_VERSION);
         version.setFrozen(true);
         scannerClass.setConstant("Version", version);
         RubyString id = runtime.newString("$Id$");
@@ -263,33 +264,21 @@ public class RubyStringScanner extends RubyObject {
     // MRI: strscan_do_scan
     private IRubyObject scan(ThreadContext context, IRubyObject regex, boolean succptr, boolean getstr, boolean headonly) {
         final Ruby runtime = context.runtime;
-
-        if (headonly) {
-            if (!(regex instanceof RubyRegexp)) {
-                regex = regex.convertToString();
-            }
-        } else {
-            if (!(regex instanceof RubyRegexp)) {
-                throw runtime.newTypeError("wrong argument type " + regex.getMetaClass() + " (expected Regexp)");
-            }
-        }
-
         check(context);
-
-        ByteList strBL = str.getByteList();
-        int strBeg = strBL.getBegin();
-
         clearMatched();
 
-        if (restLen() < 0) {
+        int restLen = restLen();
+        if (restLen < 0) {
             return context.nil;
         }
 
+        ByteList strBL = str.getByteList();
+        int currPtr = strBL.getBegin() + curr;
+
         if (regex instanceof RubyRegexp) {
             pattern = ((RubyRegexp) regex).preparePattern(str);
 
-            int currPtr = currPtr();
-            int range = currPtr + restLen();
+            int range = currPtr + restLen;
 
             Matcher matcher = pattern.matcher(strBL.getUnsafeBytes(), matchTarget(), range);
             final int ret;
@@ -311,22 +300,26 @@ public class RubyStringScanner extends RubyObject {
             }
             if (ret < 0) return context.nil;
         } else {
-            RubyString pattern = (RubyString) regex;
-
-            str.checkEncoding(pattern);
-
-            if (restLen() < pattern.size()) {
-                return context.nil;
-            }
-
+            RubyString pattern = regex.convertToString();
+            Encoding patternEnc = str.checkEncoding(pattern);
             ByteList patternBL = pattern.getByteList();
             int patternSize = patternBL.realSize();
 
-            if (ByteList.memcmp(strBL.unsafeBytes(), strBeg + curr, patternBL.unsafeBytes(), patternBL.begin(), patternSize) != 0) {
-                return context.nil;
+            if (headonly) {
+                if (restLen < pattern.size()) {
+                    return context.nil;
+                }
+                if (ByteList.memcmp(strBL.unsafeBytes(), currPtr, patternBL.unsafeBytes(), patternBL.begin(), patternSize) != 0) {
+                    return context.nil;
+                }
+                setRegisters(patternSize);
+            } else {
+                int pos = StringSupport.index(strBL, patternBL, currPtr, patternEnc);
+                if (pos == -1) {
+                    return context.nil;
+                }
+                setRegisters(patternSize + pos - curr);
             }
-
-            setRegisters(patternSize);
         }
 
         setMatched();
@@ -511,6 +504,34 @@ public class RubyStringScanner extends RubyObject {
         return get_byte(context);
     }
 
+    @JRubyMethod(name = "scan_byte")
+    public IRubyObject scan_byte(ThreadContext context) {
+        Ruby runtime = context.runtime;
+        check(context);
+        clearMatched();
+        ByteList byteList = str.getByteList();
+        int curr = this.curr;
+        if (curr >= byteList.getRealSize()) return context.nil;
+
+        int bite = byteList.get(curr);
+        prev = curr;
+        this.curr++;
+
+        setMatched();
+        adjustRegisters();
+        return RubyFixnum.newFixnum(runtime, bite);
+    }
+
+    @JRubyMethod(name = "peek_byte")
+    public IRubyObject peek_byte(ThreadContext context) {
+        check(context);
+        ByteList byteList = str.getByteList();
+        int curr = this.curr;
+        if (curr >= byteList.getRealSize()) return context.nil;
+
+        return RubyFixnum.newFixnum(context.runtime, byteList.get(curr));
+    }
+
     @JRubyMethod(name = "peek")
     public IRubyObject peek(ThreadContext context, IRubyObject length) {
         check(context);
@@ -537,6 +558,100 @@ public class RubyStringScanner extends RubyObject {
         return peek(context, length);
     }
 
+    @JRubyMethod(name = "scan_base10_integer", visibility = PRIVATE)
+    public IRubyObject scan_base10_integer(ThreadContext context) {
+        final Ruby runtime = context.runtime;
+        check(context);
+        clearMatched();
+
+        strscanMustAsciiCompat(runtime);
+
+        ByteList bytes = str.getByteList();
+        int ptr = curr;
+        int len = 0;
+
+        int remaining_len = bytes.realSize() - curr;
+
+        if (remaining_len <= 0) {
+            return context.nil;
+        }
+
+        if (bytes.get(ptr + len) == '-' || bytes.get(ptr + len) == '+') {
+            len++;
+        }
+
+        if (!Character.isDigit(bytes.get(ptr + len))) {
+            return context.nil;
+        }
+
+        setMatched();
+        prev = ptr;
+
+        while (len < remaining_len && Character.isDigit(bytes.get(ptr + len))) {
+            len++;
+        }
+
+        return strscanParseInteger(runtime, bytes, ptr, len, 10);
+    }
+
+    @JRubyMethod(name = "scan_base16_integer", visibility = PRIVATE)
+    public IRubyObject scan_base16_integer(ThreadContext context) {
+        final Ruby runtime = context.runtime;
+        check(context);
+        clearMatched();
+
+        strscanMustAsciiCompat(runtime);
+
+        ByteList bytes = str.getByteList();
+        int ptr = this.curr;
+
+        int remaining_len = bytes.realSize() - ptr;
+
+        if (remaining_len <= 0) {
+            return context.nil;
+        }
+
+        int len = 0;
+
+        if (bytes.get(ptr + len) == '-' || bytes.get(ptr + len) == '+') {
+            len++;
+        }
+
+        if ((remaining_len >= (len + 2)) && bytes.get(ptr + len) == '0' && bytes.get(ptr + len + 1) == 'x') {
+            len += 2;
+        }
+
+        if (len >= remaining_len || !isHexChar(bytes.get(ptr + len))) {
+            return context.nil;
+        }
+
+        setMatched();
+        adjustRegisters();
+        prev = ptr;
+
+        while (len < remaining_len && isHexChar(bytes.get(ptr + len))) {
+            len++;
+        }
+
+        return strscanParseInteger(runtime, bytes, ptr, len, 16);
+    }
+
+    private RubyInteger strscanParseInteger(Ruby runtime, ByteList bytes, int ptr, int len, int base) {
+        this.curr = ptr + len;
+
+        return ConvertBytes.byteListToInum(runtime, bytes, ptr, len, base, true);
+    }
+
+    private void strscanMustAsciiCompat(Ruby runtime) {
+        if (!str.getEncoding().isAsciiCompatible()) {
+            throw runtime.newEncodingCompatibilityError("ASCII incompatible encoding: " + str.getEncoding());
+        }
+    }
+
+    private static boolean isHexChar(int c) {
+        return Character.isDigit(c) || ('a' <= c && c <= 'f') || ('A' <= c && c <= 'F');
+    }
+
     @JRubyMethod(name = "unscan")
     public IRubyObject unscan(ThreadContext context) {
         check(context);
